--- Post-processing, example #1.

--
-- Permission is hereby granted, free of charge, to any person obtaining
-- a copy of this software and associated documentation files (the
-- "Software"), to deal in the Software without restriction, including
-- without limitation the rights to use, copy, modify, merge, publish,
-- distribute, sublicense, and/or sell copies of the Software, and to
-- permit persons to whom the Software is furnished to do so, subject to
-- the following conditions:
--
-- The above copyright notice and this permission notice shall be
-- included in all copies or substantial portions of the Software.
--
-- THE SOFTWARE IS PROVIDED "AS IS", WITHOUT WARRANTY OF ANY KIND,
-- EXPRESS OR IMPLIED, INCLUDING BUT NOT LIMITED TO THE WARRANTIES OF
-- MERCHANTABILITY, FITNESS FOR A PARTICULAR PURPOSE AND NONINFRINGEMENT.
-- IN NO EVENT SHALL THE AUTHORS OR COPYRIGHT HOLDERS BE LIABLE FOR ANY
-- CLAIM, DAMAGES OR OTHER LIABILITY, WHETHER IN AN ACTION OF CONTRACT,
-- TORT OR OTHERWISE, ARISING FROM, OUT OF OR IN CONNECTION WITH THE
-- SOFTWARE OR THE USE OR OTHER DEALINGS IN THE SOFTWARE.
--
-- [ MIT license: http://www.opensource.org/licenses/mit-license.php ]
--
display.setDefault("isShaderCompilerVerbose", true)
-- Boilerplate for distortion-based shaders.
local function DistortionBindCanvasEffect (object, fill, name)
	object.fill = fill
	object.fill.effect = name
	object.fill.effect.xdiv = 1 / display.contentWidth
	object.fill.effect.ydiv = 1 / display.contentHeight
end

local function DistortionKernelParams ()
	return {
		{ index = 0, name = "xdiv" },
		{ index = 1, name = "ydiv" },
		{ index = 2, name = "alpha", default = 1, min = 0, max = 1 }
	}
end

-- An effect that shimmers a texture, in this example the captured background.
do
	local kernel = { language = "glsl", category = "filter", name = "shimmer" }

	kernel.vertexData = DistortionKernelParams()

	kernel.vertexData[4] = { name = "influence", index = 3, min = .5, max = 1024., default = 15. }

	kernel.vertex = [[
		varying P_POSITION vec2 v_Pos;

		P_POSITION vec2 VertexKernel (P_POSITION vec2 pos)
		{
			v_Pos = pos;

			return pos;
		}
	]]

	kernel.fragment = [[
		varying P_POSITION vec2 v_Pos;

		P_COLOR vec3 GetDistortedRGB (sampler2D s, P_UV vec2 offset, P_UV vec3 divs_alpha)
		{
			P_UV vec2 uv = (v_Pos + offset) * divs_alpha.xy;

			return texture2D(s, uv).rgb * divs_alpha.z;
		}

		P_COLOR vec3 GetDistortedRGB (sampler2D s, P_UV vec2 offset, P_UV vec4 divs_alpha)
		{
			return GetDistortedRGB(s, offset, divs_alpha.xyz);
		}

		P_UV float GetDistortInfluence (P_UV vec2 uvn, P_UV float low, P_UV float scale)
		{
			return (1. - smoothstep(low, 1., dot(uvn, uvn))) * scale;
		}

		P_POSITION float hash1 (P_UV float n)
		{
		#if !defined(GL_FRAGMENT_PRECISION_HIGH) || GL_FRAGMENT_PRECISION_HIGH
			return fract(sin(n) * 4375.85453);
		#else
			return fract(sin(n) * 43.7585453);
		#endif
		}

		P_POSITION float IQ (P_POSITION vec2 x)
		{
			P_POSITION vec2 p = floor(x);
			P_POSITION vec2 f = fract(x);

			f = f * f * (3.0 - 2.0 * f);

			P_POSITION float n = p.x + p.y * 57.0;

			return mix(mix(hash1(n +  0.0), hash1(n +  1.0), f.x),
					mix(hash1(n + 57.0), hash1(n + 58.0), f.x), f.y);
		}

		P_POSITION vec2 IQ_Octaves (P_POSITION vec2 x, P_POSITION vec2 y)
		{
			return vec2(IQ(x) * .5, IQ(y) * .25);
		}

		P_COLOR vec4 FragmentKernel (P_UV vec2 uv)
		{
			P_UV vec2 uvn = 2. * uv - 1.;
			P_UV vec2 offset = IQ_Octaves(uv * 14.1, uv * 12.3) * GetDistortInfluence(uvn, .75, CoronaVertexUserData.w);
			P_COLOR vec3 background = GetDistortedRGB(CoronaSampler0, offset, CoronaVertexUserData);

			return CoronaColorScale(vec4(background, 1.));
		}
	]]

	graphics.defineEffect(kernel)
end

-- Vlad's version of flame: https://forums.coronalabs.com/topic/56284-share-your-shaders/?p=291856
do
	local kernel = { category = "generator", group = "vlad", name = "flame" }

	kernel.vertexData = {
		{ name = "count", index = 0, default = 5 },
		{ name = "height", index = 1, default = 4 },
		{ name = "after_burn", index = 2, default = 2.5 }
	}
	
	kernel.isTimeDependent = true

	kernel.fragment = [[
		P_DEFAULT vec2 hash( P_DEFAULT vec2 p )
		{
			p = vec2( dot(p,vec2(127.1,311.7)),
					 dot(p,vec2(269.5,183.3)) );
			return -1.0 + 2.0*fract(sin(p)*43758.5453123);
		}

		P_DEFAULT float noise( in P_DEFAULT vec2 p )
		{
			const P_DEFAULT float K1 = 0.366025404; // (sqrt(3)-1)/2;
			const P_DEFAULT float K2 = 0.211324865; // (3-sqrt(3))/6;
			P_DEFAULT vec2 i = floor( p + (p.x+p.y)*K1 );
			P_DEFAULT vec2 a = p - i + (i.x+i.y)*K2;
			P_DEFAULT vec2 o = (a.x>a.y) ? vec2(1.0,0.0) : vec2(0.0,1.0);
			P_DEFAULT vec2 b = a - o + K2;
			P_DEFAULT vec2 c = a - 1.0 + 2.0*K2;
			P_DEFAULT vec3 h = max( 0.5-vec3(dot(a,a), dot(b,b), dot(c,c) ), 0.0 );
			P_DEFAULT vec3 n = h*h*h*h*vec3( dot(a,hash(i+0.0)), dot(b,hash(i+o)), dot(c,hash(i+1.0)));
			return dot( n, vec3(70.0) );
		}

		P_DEFAULT float fbm(P_DEFAULT vec2 uv)
		{
			P_DEFAULT float f;
			P_DEFAULT mat2 m = mat2( 1.6,  1.2, -1.2,  1.6 );
			f  = 0.5000*noise( uv ); uv = m*uv;
			f += 0.2500*noise( uv ); uv = m*uv;
			f += 0.1250*noise( uv ); uv = m*uv;
			f += 0.0625*noise( uv ); uv = m*uv;
			f = 0.5 + 0.5*f;
			return f;
		}

		// no defines, standard redish flames
		//#define BLUE_FLAME
		//#define GREEN_FLAME

		P_COLOR vec4 FragmentKernel( P_UV vec2 texCoord )
		{
			P_UV vec2 uv = texCoord.xy;
			P_UV vec2 q = uv;
			q.x *= CoronaVertexUserData.x;
			q.y *= CoronaVertexUserData.y;
			P_DEFAULT float strength = floor(q.x+1.);
			P_DEFAULT float T3 = max(3.,1.25*strength)*CoronaTotalTime;
			q.x = mod(q.x,CoronaVertexUserData.x)-CoronaVertexUserData.x/2.;
			q.y -= 0.25;
			P_DEFAULT float n = fbm(strength*q - vec2(0,T3));
			P_DEFAULT float c = 1. - 16. * pow( max( 0., length(q*vec2(1.8+q.y*1.5,.75) ) - n * max( 0., q.y+.25 ) ),1.2 );
			P_DEFAULT	float c1 = n * c * (1.5-pow(CoronaVertexUserData.z*uv.y,4.));
			c1=clamp(c1,0.,1.);
			P_COLOR vec3 col = vec3(1.5*c1, 1.5*c1*c1*c1, c1*c1*c1*c1*c1*c1);
			#ifdef BLUE_FLAME
				col = col.zyx;
			#endif
			#ifdef GREEN_FLAME
				col = 0.85*col.yxz;
			#endif
			P_DEFAULT float a = c * (1.-pow(uv.y,3.));
			return CoronaColorScale(vec4( mix(vec3(0.),col,a), a));
		}
	]]

	graphics.defineEffect(kernel)
end

-- Save some typing.
local CX, W = display.contentCenterX, display.contentWidth
local CY, H = display.contentCenterY, display.contentHeight

-- Add an image and a square in a background group.
local Background = display.newGroup()
local Image = display.newImageRect(Background, "Image.jpg", W, H)

Image:translate(CX, CY)

local Square = display.newRect(Background, CX, CY, 150, 250)

Square:setFillColor(0, 0, 1)

-- Rig up a canvas that that covers the screen, capturing certain layers for use in
-- post-processing.
local Canvas = graphics.newTexture{
	type = "canvas", width = W, height = H
}
		
Canvas.anchorX, Canvas.anchorY = -.5, -.5

Canvas:draw(Background) -- in particular, these layers...

Runtime:addEventListener("enterFrame", function()
	Canvas:invalidate("cache")
end) -- ...to be invalidated by this each frame

-- Give it a frame to take hold, and finish up.
timer.performWithDelay(50, function()
	-- The canvas needs to be on something to be seen, so put it on a full-screen rect.
	local CanvasRect = display.newImageRect(Canvas.filename, Canvas.baseDir, W, H)

	CanvasRect.x, CanvasRect.y = CX, CY

	-- Put fire (just for looks) and a stream of heat haze particles above the background.
	-- Note that these heat haze particles are NOT transparent, despite the way they make
	-- things look, thus the fire follows the haze layer.
	local haze_layer = display.newGroup()
	local fire = display.newRect(CX, CY, W, H)  -- the effect itself is full-screen, despite appearances,
												-- sort of a side effect of its Shadertoy beginnings

	fire.yScale = -1

	fire.fill.effect = "generator.vlad.flame"

	-- Slide the image back and forth to show that it's picking up the screen, not just using
	-- the image as a texture.
	Background.to_x = CX / 2
	Background.time = 3500

	local move_params = {}

	local function Move (object)
		move_params.x = object.to_x
		move_params.time = object.time

		transition.to(object, move_params)

		object.to_x = -object.to_x
	end
	
	move_params.onComplete = Move

	Move(Background)

	-- Move the square too, a little differently than the image, to accentuate this.
	Square.to_x = 500
	Square.time = 7000

	Move(Square)

	-- Create an endless stream of haze objects in the layer beneath the fire.
	local Fill = { type = "image", filename = Canvas.filename, baseDir = Canvas.baseDir }

	timer.performWithDelay(80, function()
		local haze = display.newRect(haze_layer, CX + math.random(-10, 10), H - 50 - math.random(30), 80, 80)

		DistortionBindCanvasEffect(haze, Fill, "filter.custom.shimmer") -- comment out this line to see the raw objects

		transition.to(haze, {
			y = haze.y - 450,
			time = math.random(1500, 3500),
			rotation = math.random(-3000, 3000),
			onComplete = display.remove
		})
	end, 0)
end)