--- Post-processing, example #2.

--
-- Permission is hereby granted, free of charge, to any person obtaining
-- a copy of this software and associated documentation files (the
-- "Software"), to deal in the Software without restriction, including
-- without limitation the rights to use, copy, modify, merge, publish,
-- distribute, sublicense, and/or sell copies of the Software, and to
-- permit persons to whom the Software is furnished to do so, subject to
-- the following conditions:
--
-- The above copyright notice and this permission notice shall be
-- included in all copies or substantial portions of the Software.
--
-- THE SOFTWARE IS PROVIDED "AS IS", WITHOUT WARRANTY OF ANY KIND,
-- EXPRESS OR IMPLIED, INCLUDING BUT NOT LIMITED TO THE WARRANTIES OF
-- MERCHANTABILITY, FITNESS FOR A PARTICULAR PURPOSE AND NONINFRINGEMENT.
-- IN NO EVENT SHALL THE AUTHORS OR COPYRIGHT HOLDERS BE LIABLE FOR ANY
-- CLAIM, DAMAGES OR OTHER LIABILITY, WHETHER IN AN ACTION OF CONTRACT,
-- TORT OR OTHERWISE, ARISING FROM, OUT OF OR IN CONNECTION WITH THE
-- SOFTWARE OR THE USE OR OTHER DEALINGS IN THE SOFTWARE.
--
-- [ MIT license: http://www.opensource.org/licenses/mit-license.php ]
--

-- Copies part of the screen into the rect.
do
	local kernel = { category = "filter", name = "fetch_screen_pixels" }

	kernel.vertexData = {
		{ name = "u0", index = 0, default = 0 },
		{ name = "v0", index = 1, default = 0 },
		{ name = "u1", index = 2, default = 1 },
		{ name = "v1", index = 3, default = 1 }
	}

	kernel.vertex = [[
		P_POSITION vec2 VertexKernel (P_POSITION vec2 pos)
		{
			v_TexCoord = mix(CoronaVertexUserData.xy, CoronaVertexUserData.zw, CoronaTexCoord.xy);

			return pos;
		}
	]]

	graphics.defineEffect(kernel)
end

-- Piggyback various Corona filters on the screen capture
local function TwoPass (name, new_name)
	local kernel = { category = "filter", name = new_name }

	kernel.graph = {
		nodes = {
			fetch = { effect = "filter.custom.fetch_screen_pixels", input1 = "paint1" },
			corona_filter = { effect = "filter." .. name, input1 = "fetch" }, -- can add input2 = "paint2" for composite
																			  -- effects (changing some names, of course)
		},
		output = "corona_filter"
	}

	graphics.defineEffect(kernel)
end

--
local Effects = {}

for _, name in ipairs{
	"bloom",
	"blur",
	"crystallize",
	"emboss",
	"frostedGlass",
	"grayscale",
	"invert",
	"pixelate",
	"posterize",
	"scatter",
	"sepia",
	"sobel",
	"vignette",
	"wobble",
	"woodCut",
	"zoomBlur"
} do
	TwoPass(name, name .. "_twopass")

	Effects[#Effects + 1] = name
end

-- Save some typing.
local CX, W = display.contentCenterX, display.contentWidth
local CY, H = display.contentCenterY, display.contentHeight

-- Add an image in the background group.
local Background = display.newGroup()
local Image = display.newImageRect(Background, "Image.jpg", W, H)

Background:translate(CX, CY)

-- Rig up a canvas that that covers the screen, capturing certain layers for use in
-- post-processing.
local Canvas = graphics.newTexture{
	type = "canvas", width = W, height = H
}
		
Canvas.anchorX, Canvas.anchorY = -.5, -.5

Canvas:draw(Background) -- in particular, these layers...

Runtime:addEventListener("enterFrame", function()
	Canvas:invalidate("cache")
end) -- ...to be invalidated by this each frame

-- Give it a frame to take hold, and finish up.
timer.performWithDelay(50, function()
	-- The canvas needs to be on something to be seen, so put it on a full-screen rect.
	local CanvasRect = display.newImageRect(Canvas.filename, Canvas.baseDir, W, H)

	CanvasRect.x, CanvasRect.y = CX, CY

	-- Slide an effect window back and forth (actually, we do this indirectly through a
	-- proxy, see below). Make this invisible to start with while we wait for the timer
	-- to elapse for the first time.
	local window = display.newRect(CX, CY, 200, H)

	window:setStrokeColor(0)
	
	window.strokeWidth = 2
	window.isVisible = false

	-- Multi-pass shaders get put into an offscreen buffer, so their geometry is basically
	-- useless for figuring out where we are on the screen. Therefore, we need to pass this
	-- information in explicitly, and we might as well just calculate the texture coordinates
	-- instead. (Since in this example our rect occupies the full screen height, the defaults
	-- are already the proper values.) The proxy lets us capture the transition's changes and
	-- track the window slightly better than an "enterFrame" listener.
	local proxy = {}

	proxy.dx = (W - window.width) / 2
	proxy.time = 2500

	local function SetParams (effect, x)
		effect.fetch.u0 = (x - window.width / 2) / W
		effect.fetch.u1 = (x + window.width / 2) / W
	end
local ee
	setmetatable(proxy, {
		__index = function(_, k)
			return window[k]
		end,
		__newindex = function(_, k, v)
			window[k] = v

--			local effect = window.fill.effect

			if ee--[[effect]] and k == "x" then
				window.fill.effect = ee
				SetParams(window.fill.effect, v)
			end
		end
	})

	local move_params = {}

	local function Move (object)
		move_params.x = CX + object.dx
		move_params.time = object.time

		transition.to(object, move_params)

		object.dx = -object.dx
	end
	
	move_params.onComplete = Move

	Move(proxy)

	-- Cycle through the effects.
	local fill = { type = "image", filename = Canvas.filename, baseDir = Canvas.baseDir }
	local name = display.newText("No Effect", CX, H - 50, native.systemFontBold, 50)

	name:setFillColor(1, 0, 0)

	timer.performWithDelay(1500, function(event)
		local index = (event.count - 1) % #Effects + 1

		name.text = Effects[index]

		window.isVisible = true -- make visible the first time
		window.fill = fill
		ee = "filter.custom." .. Effects[index] .. "_twopass"
		window.fill.effect = ee

		SetParams(window.fill.effect, window.x)
	end, 0)
end)